/** @file memory.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ memory ]
*/

#ifndef _MEMORY_H_
#define _MEMORY_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"
#include "../../config.h"
#include "../hd46505.h"
#include "sound.h"
#include "display.h"
#include "comm.h"
#include "cmt.h"
#include "pia.h"
#include "via.h"
#include "../event.h"

#define SIG_MEMORY_DISP		0
#define SIG_MEMORY_VSYNC	1
//#define SIG_MEMORY_PIA_PA	2
#define SIG_TRACE_COUNTER	4
#define SIG_TRACE_COUNTER_STOP	5

#define S1_BANK_SIZE	11	///< 2Kbytes
#define L3_BANK_SIZE	4	///< 16bytes


/**
	@brief Memory access
*/
class MEMORY : public DEVICE
{
private:
	DEVICE *d_cpu, *d_pia, *d_crtc, *d_disp, *d_acia;
	DEVICE *d_key, *d_mouse, *d_sound, *d_comm, *d_cmt, *d_timer;
	DEVICE *d_kanji, *d_fdc3, *d_fdc5, *d_fdc5_ex, *d_fdd;
	DEVICE *d_psgst, *d_psgex, *d_psg3, *d_psg9, *d_fmopn;
	DEVICE *d_pia_ex, *d_acia_ex, *d_comm1, *d_board;
	DEVICE *d_event;

	uint8_t l3rom[0x6000];		///< 24KB
	uint8_t s1rom[0x10000];		///< 64KB
	uint8_t s1dicrom[0x8000];		///< 32KB
	uint8_t rom1802[0x800];
	uint8_t rom1805[0x800];
	uint8_t *cm01rom;				///< 32KB
	uint8_t *cm01eeprom;			///< 8KB

	uint8_t ram[0x10000];			///< main ram (S1 is also used as graphic vram red)
//	uint8_t color_ram[0x4000];	///< 0x0400 - 0x43ff -> use vgram
	uint8_t ig_ram[0x0800 * 3];	///< 0xa000 - 0xa7ff * 3 (b,r,g)
//	uint8_t ram2[0x10000];		///< extended ram
	int   exram_size;			///< extended ram size (KB)
	uint8_t *exram;				///< extended ram (variable)

	uint8_t vgram[0x8000];		///< S1 graphic ram (green,blue)(32KB)
	uint8_t vtram[0x800];			///< S1 text ram (2KB)
	uint8_t vcram[0x800];			///< S1 color ram (2KB)
	uint8_t *font;				///< font rom (4KB)

	uint8_t addr_map[64][16];		///< address mapping table

	uint8_t wdmy[1 << S1_BANK_SIZE];
	uint8_t rdmy[1 << S1_BANK_SIZE];
	uint8_t* l3wbank[0x10000 >> L3_BANK_SIZE];
	uint8_t* l3rbank[0x10000 >> L3_BANK_SIZE];
	uint8_t l3banktype[0x10000 >> L3_BANK_SIZE];
	uint8_t* s1wbank[0x100000 >> S1_BANK_SIZE];
	uint8_t* s1rbank[0x100000 >> S1_BANK_SIZE];
	uint8_t s1banktype[0x100000 >> S1_BANK_SIZE];

	/// bit0:read wait bit4:write wait
	/// bit1,5: refresh memory wait
	/// bit2,6: disptmg wait
	uint8_t  l3wait[0x10000 >> L3_BANK_SIZE];
	uint8_t  s1wait[0x100000 >> S1_BANK_SIZE];

	uint8_t color_reg;		///< 0xffd8

//	uint8_t ig_enable;		///< 0xffe9
	uint8_t ig_enreg;			///< 0xffea

	uint8_t mem_bank_reg1;	///< 0xffc0
	uint8_t mem_bank_reg2;	///< 0xffe8
	bool  mem_bank_mask;	///< 0xffc0 mask

//	uint8_t vram_sel;
//	uint8_t cpu_wait;			///< io access (bit1) 0:1MHz 2:2MHz

//	int   trace_counter_id;
//	uint32_t pc, pc_prev;
	int   trace_counter_phase;
	int   trace_count_remain;
	int   prev_trace_clock;

	bool  l3rom_loaded;
	bool  s1rom1_loaded;
	bool  s1rom2_loaded;
	bool  s1dicrom_loaded;
	bool  rom1802_loaded;
	bool  rom1805_loaded;
	bool  rom_loaded_at_first;

	bool  cm01rom_loaded;
	bool  cm01eeprom_loaded;

	bool  cm01rom_access_ok;
	bool  cm01basic_enable;
	uint8_t cm01eeprom_bank;
	bool  cm01eeprom_writing;

	uint32_t addr_bank;
	uint32_t addr_seg04, addr_seg16, addr_seg32;
	uint32_t addr_comio, addr_64kb;

	bool  bfxxx_access;

#ifdef USE_RTC
	uint8_t rtc_reg_sel;
	uint8_t rtc_buf;
#endif

	//for resume
#pragma pack(1)
	struct vm_state_st_l3 {
		uint8_t ram[0x10000];
		uint8_t ram2[0x10000];		// extended ram
		uint8_t color_ram[0x4000];	// 0x0400 - 0x43ff
		uint8_t ig_ram[0x0800 * 3];	// 0xa000 - 0xa7ff * 3 (b,r,g)

		uint8_t mem_bank_reg1;	// from PIA port A
		uint8_t mem_bank_reg2;	// from 0xffe8

		int   trace_counter_id;
		uint32_t pc;
		uint32_t pc_prev;

		// from config
		uint8_t dipswitch;
		int fdd_type;
		int io_port;

		char  reserved[9];
	};

	struct vm_state_st_s1 {
		uint8_t ram[0x10000];
		uint8_t vgram[0x8000];		// graphic ram
		uint8_t vtram[0x800];			// text chr
		uint8_t vcram[0x800];			// text color
		uint8_t ig_ram[0x0800 * 3];	// ig (b,r,g)

//		int   trace_counter_id;
//		uint32_t pc;
//		uint32_t pc_prev;
		int   trace_counter_phase;	// version 0x42
		int   trace_count_remain;	// version 0x42
		int   prev_trace_clock;		// version 0x42

		uint8_t mem_bank_reg1;	// from 0xffc0
		uint8_t mem_bank_reg2;	// from 0xffe8
		uint8_t mem_bank_mask;	// from 0xffc0 mask
		char  reserved1;

		// from config
		uint8_t dipswitch;
		uint8_t sys_mode;
		int fdd_type;
		int io_port;

		uint8_t mode_sel;
		uint8_t color_reg;
		uint8_t ig_enreg;
		uint8_t reg_fuse;
		uint8_t reg_trapf;
		uint8_t reg_busctrl;

		union {
			struct {
				uint8_t addr_map[16];
				char  reserved2[256];	// for OS-9 exboard

				char  exram_ident[14];
				uint16_t exram_size;
				// exram (variable)
				// uint8_t exram[0 - 512KB]
			} v42;
			struct {
				uint8_t addr_map[64][16];
				uint8_t reg_addrseg;

				char reserved3[7];
				char exram_ident[6];
				uint16_t exram_size;
				// exram (variable)
				// uint8_t exram[0 - 512KB]
			} v43;
		};
	};
#pragma pack()

	void SET_BANK(int s, int e, uint8_t* w, uint8_t* r, uint8_t** wbank, uint8_t** rbank, int bank_size, uint8_t *waitbank, uint8_t waitval, uint8_t *banktype, uint8_t type);

	void load_rom_files();
	void save_rom_files();
	void reset_sram(uint8_t *sram, size_t size);
	void reset_igram();
	void reset_s1_membank();
	void reset_cm01_membank();
	void reset_l3_membank();
	void reset_trace_counter();
	void set_igmode(uint32_t data);
	inline void process_trace_counter(int clk);
	inline void fetch_trace_counter(int clk);
	inline void set_trace_counter(int clk);
	inline void fire_trace_counter(int clk);
	inline void clear_trace_counter(int clk);

	void change_cm01_membank();
	void change_l3_memory_bank();
	void change_system_mode(bool usermode);
	void reset_s1_l3_mode();
	inline void set_cpu_clock(uint8_t data);

	void set_extended_ram();

public:
	MEMORY(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier)
	{
		set_class_name("MEMORY");
	}
	~MEMORY() {}

	// common functions
	void initialize();
	void reset();
	void release();
	void write_data8w(uint32_t addr, uint32_t data, int *wait);
	uint32_t read_data8w(uint32_t addr, int *wait);
	void latch_address(uint32_t addr, int *wait);

//	void write_data16(uint32_t addr, uint32_t data) {
//		write_data8(addr, data & 0xff); write_data8(addr + 1, data >> 8);
//	}
//	uint32_t read_data16(uint32_t addr) {
//		return read_data8(addr) | (read_data8(addr + 1) << 8);
//	}
	void write_io8(uint32_t addr, uint32_t data);
	void write_signal(int id, uint32_t data, uint32_t mask);
	uint32_t read_signal(int id);

	// unique function
	void set_context_cpu(DEVICE* device) {
		d_cpu = device;
	}
	void set_context_pia(DEVICE* device) {
		d_pia = device;
	}
	void set_context_crtc(DEVICE* device) {
		d_crtc = device;
	}
	void set_context_display(DEVICE* device) {
		d_disp = device;
	}
	void set_context_acia(DEVICE* device) {
		d_acia = device;
	}
	void set_context_key(DEVICE* device) {
		d_key = device;
	}
	void set_context_mouse(DEVICE* device) {
		d_mouse = device;
	}
	void set_context_sound(DEVICE* device) {
		d_sound = device;
	}
	void set_context_psg(DEVICE* device0, DEVICE* device1) {
		d_psgst = device0;
		d_psgex = device1;
	}
	void set_context_psg3(DEVICE* device) {
		d_psg3 = device;
	}
	void set_context_psg9(DEVICE* device) {
		d_psg9 = device;
	}
	void set_context_fmopn(DEVICE* device) {
		d_fmopn = device;
	}
	void set_context_comm(DEVICE* device) {
		d_comm = device;
	}
	void set_context_cmt(DEVICE* device) {
		d_cmt = device;
	}
	void set_context_timer(DEVICE* device) {
		d_timer = device;
	}
	void set_context_kanji(DEVICE* device) {
		d_kanji = device;
	}
	void set_context_fdc(DEVICE* device3, DEVICE* device5, DEVICE* device5e) {
		d_fdc3 = device3;
		d_fdc5 = device5;
		d_fdc5_ex = device5e;
	}
	void set_context_fdd(DEVICE* device) {
		d_fdd = device;
	}
	void set_context_pia_ex(DEVICE* device) {
		d_pia_ex = device;
	}
	void set_context_acia_ex(DEVICE* device) {
		d_acia_ex = device;
	}
	void set_context_comm1(DEVICE* device) {
		d_comm1 = device;
	}
	void set_context_board(DEVICE* device) {
		d_board = device;
	}
	void set_context_event(DEVICE* device) {
		d_event = device;
	}
	uint8_t* get_l3vram();
	uint8_t* get_s1vtram();
	uint8_t* get_s1vcram();
	uint8_t* get_s1vgrram();
	uint8_t* get_s1vggram();
	uint8_t* get_s1vgbram();

	uint8_t* get_color_ram() {
		return vgram;
	}
	uint8_t* get_ig_ram() {
		return ig_ram;
	}
	void set_font(uint8_t *val) {
		font = val;
	}

	uint32_t address_mapping(uint32_t addr);

//	void event_callback(int event_id, int err);
	void update_config();

	void save_state(FILEIO *fio);
	void load_state(FILEIO *fio);


#ifdef USE_DEBUGGER
	uint32_t debug_latch_address(uint32_t addr);

	void debug_write_data8(int type, uint32_t addr, uint32_t data);
	uint32_t debug_read_data8(int type, uint32_t addr);
	void debug_write_data16(int type, uint32_t addr, uint32_t data);
	uint32_t debug_read_data16(int type, uint32_t addr);
	void debug_write_data32(int type, uint32_t addr, uint32_t data);
	uint32_t debug_read_data32(int type, uint32_t addr);

	uint32_t debug_physical_addr_mask(int type);

	uint32_t debug_read_bank(uint32_t addr);
	void debug_memory_map_info(DebuggerConsole *dc);
	void print_memory_bank(uint32_t data, bool w, _TCHAR *buffer, size_t buffer_len);
	void debug_address_map_info(DebuggerConsole *dc, int index);
	int debug_address_map_edit(DebuggerConsole *dc, int index, int *values, int count);
	uint32_t debug_address_mapping_rev(uint32_t addr);
	uint32_t debug_address_mapping_basic(uint32_t addr);
	void debug_memory_space_map_info(DebuggerConsole *dc, int index);
	void debug_memory_space_map_mmptbl(DebuggerConsole *dc, int num);
	void debug_memory_space_map_current(DebuggerConsole *dc, int index);
	void debug_memory_space_map_mrasgn(DebuggerConsole *dc);
	void debug_memory_space_map_mchtbl(DebuggerConsole *dc, int num);
	void debug_memory_space_map_syscall(DebuggerConsole *dc, int num);
	void debug_memory_space_map_mmdpg(DebuggerConsole *dc, int num);
	int debug_memory_space_map_edit(DebuggerConsole *dc, int index, int *values, int count);
	void debug_edit_memory_space_map_mmptbl(DebuggerConsole *dc, int num, int *values);
	void debug_edit_memory_space_map_mchtbl(DebuggerConsole *dc, int num, int *values);
	void debug_edit_memory_space_map_mmdpg(DebuggerConsole *dc, int num, int *values);

	bool debug_write_reg(uint32_t reg_num, uint32_t data);
	void debug_regs_info(_TCHAR *buffer, size_t buffer_len);

	bool get_debug_graphic_memory_size(int type, int *width, int *height);
	bool debug_draw_graphic(int type, int width, int height, scrntype *buffer);
#endif

};

#endif
